#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <string.h>
#include <strings.h>

#define ZCHAFF_LOC	"/home/nmousavi/Nima/Zchaff/zchaff"
//#define ZCHAFF_LOC	"/home/tripunit/Documents/zchaff/zchaff/zchaff"
#define ROLES_OUTPUT	0
#define KEEP_MAIN	0

void usage(char *cmd) {
    printf("Usage: %s input-file-for-zchaff\n", cmd);
}

/* invokeZchaff -- invokes zchaff with input cnfInputFile
 * Return 1 if zchaff says that it is satisfiable
 * Returns 0 if zchaff says that it is unsatisfiable
 * Return -1 on error (e.g., if zchaff returns an error.
 *
 * The run time as reported by zchaff is stored in runTime on return.
 * Of course it is not meaningful is the return is -1.
 */
int invokeZchaff(char *cnfInputFile, double &runTime) {
    int pfd[2]; /* pfd[0] is read-end, pfd[1] is write-end */

    if(pipe(pfd) < 0) {
	perror("pipe"); exit(EXIT_SUCCESS);
    }

    if(!cnfInputFile) {
	printf("invokeZchaff::cnfInputFile is NULL!\n");
	return -1;
    }

    if(fork() == 0) {
	/* Child */
	close(pfd[0]); /* Nothing to read from pipe */

	if(dup2(pfd[1], STDOUT_FILENO) < 0) {
	    perror("dup2 - stdout"); exit(EXIT_SUCCESS);
	}

	if(dup2(pfd[1], STDERR_FILENO) < 0) {
	    perror("dup2 - stderr"); exit(EXIT_SUCCESS);
	}

	if(execl(ZCHAFF_LOC, ZCHAFF_LOC, cnfInputFile, (char *)NULL) < 0) {
	    perror("execl"); exit(EXIT_SUCCESS);
	}
    }
    else {
	/* Parent */
	close(pfd[1]); /* Nothing to write to pipe */

	int readRet = 0;
	const int strSize = strlen("Satisfiable") + 1;
	char c, satStr[strSize];
	int satUnsatFound = -1; /* -1: neither
			            0: Sat found
				    1: Unsat found
				*/
	int foundRunTime = 0, foundDecimal = 0;
	const int runTimeStrLen = 128;
	char runTimeStr[runTimeStrLen];
#if ROLES_OUTPUT
	int thisVar = 0;
	int doneVars = 0;
#endif

	bzero(satStr, strSize);
	bzero(runTimeStr, runTimeStrLen);
	runTime = 0.0;

	for(readRet = read(pfd[0], (void *)&c, 1);
		readRet > 0;
	    readRet = read(pfd[0], (void *)&c, 1))
	{

	    if(satUnsatFound < 0) {
		char tmpStr[strSize];

		bzero(tmpStr, strSize);
		bcopy(satStr+1, tmpStr, strSize-2);
		tmpStr[strSize - 2] = c; tmpStr[strSize - 1] = NULL;
		bcopy(tmpStr, satStr, strSize-1);
#if 0
		printf("satStr: %s\n", satStr);
#endif

		if(!strncmp(satStr, "Unsatisfiable", strSize - 1)) {
#if 0
		    printf("Unsatisfiable.");
#endif
		    satUnsatFound = 0;
		}
		else if(!strncmp(satStr, "Satisfiable", strSize - 1)) {
#if 0
		    printf("Satisfiable.");
#endif
#if ROLES_OUTPUT
		    printf(" Variables: ");
#endif

		    satUnsatFound = 1;
		}
	    }
#if ROLES_OUTPUT
	    else if((satUnsatFound == 1) && (!doneVars)) {
		if(c == '-') {
		    thisVar = -1;
		}
		else if((c < '0') || (c > '9')) {
		    if(thisVar > 0) {
			printf("%d ", thisVar);
		    }
		    thisVar = 0;
		    if((c >= 'a' && c <= 'z') ||
		       (c >= 'A' && c <= 'Z')) {

			/* Done reading vars */
			doneVars = 1;
		    }
		}
		else {
		    thisVar = thisVar * 10 + (int)(c - '0');
		}
	    }
#endif
	    else {
		if(!foundRunTime) {
		    /* Look for run time */
		    char tmpStr[strSize];

		    bzero(tmpStr, strSize);
		    bcopy(satStr+1, tmpStr, strSize-2);
		    tmpStr[strSize-2] = c; tmpStr[strSize - 1] = NULL;
		    bcopy(tmpStr, satStr, strSize-1);

#if 0
		    printf("satStr looking for runtime: %s\n", satStr);
#endif

		    if(!strncmp(satStr+(strSize - 1 - strlen("Run Time")), "Run Time", strlen("Run Time"))) {
#if 0
			printf("satStr @ finding runtime: %s\n", satStr);
#endif
			foundRunTime = 1;
		    }
		}
		else {
#if 0
		    printf("After runtime: %c ", c);
#endif
		    if(c >= 'A' && c <= 'Z') {
			/* Done reading run time */
			sscanf(runTimeStr, "%lf", &runTime);
			break;
		    }
		    else if((c >= '0' && c <= '9') || (c == '.')) {
			char oneC[2];

			oneC[0] = c; oneC[1] = NULL;
			strcat(runTimeStr, oneC);
		    }
		    else {
			/* Ignore, but continue reading */
			continue;
		    }
		}
	    }
	}

	if(readRet < 0) {
	    perror("read"); exit(EXIT_SUCCESS);
	}

	if(satUnsatFound < 0) {
	    printf("Error -- neither sat nor unsat.");
	}

	if(!foundRunTime) {
	    printf("Could not find Run Time.");
	}

	close(pfd[0]);
	return satUnsatFound;
    }

    close(pfd[0]);
    return -1;
}

#if KEEP_MAIN
int main(int argc, char *argv[])
{
    if(argc < 2) {
	usage(argv[0]); exit(EXIT_SUCCESS);
    }

    double runTime = 0.0;

    int ret = invokeZchaff(argv[1], runTime);

    if(ret < 0) {
	printf("invokeZchaff() returned error.\n");
	exit(EXIT_SUCCESS);
    }

    if(ret) {
	printf("Satisfiable. ");
    }
    else {
	printf("Unsatisfiable. ");
    }

    printf("Run time: %lf\n", runTime);

    return 0;
}
#endif
