#ifndef SEARCH_ALGO_H
#define SEARCH_ALGO_H


#include <iostream>
#include<vector>
#include<string.h>
#include <sstream>
#include<fstream>
#include"stdlib.h"
#include<math.h>
#include<stdint.h>


using namespace std;

class Search_algo {
      
                  private:
                           char* file;
                           int Roles_indices;                    // 25  I.e., roles are 1 - 25 
                           int Perms_indices;                    // 50  I.e., perms are 26 - 75 */
                           
                           vector< vector<int> > D;              // Each Constraint is assigned to those Roles 
                           int OBJ;                      // 0 means min, 1 means max. NOTE: this is irrelevant if #PUB-PLB above has 0       
                          
                   
                  public:
                         vector< vector<int> > PR_PLB;              // Each perm (in Plb) is assigned to those roles
                         vector< vector<int> > PR_PUB_PLB;          // Each perm (Pub\Plb) is assigned to those roles
			 vector<vector<int> > PR_POUT;              //perm role assign of all roles outside of PUB
                          
                         vector<int> PLB;              // set of permissions, P_lb, that mut be activate 
                         vector<int> PUB_PLB;          //P_ub - P_lb, NOTE: this line contains 0 for P_ub -P_lb = \emptyset
                       
                         Search_algo(char*);
                         ~Search_algo();
                         
                         void build_vectors();
                         vector<int> gen_roles(int n);
                         bool Is_any(int );

                         bool PUB_PLB_OK(const vector<int> &, int) const;
                         bool D_OK(const vector<int> &) const;
			 bool POUT_OK(const vector<int> &) const;

                         bool member(const vector<int> &, const int &) const;
                         bool d_OK(const vector<int>&, const vector<int> &) const;
                         void printVector(const vector<int> &v) const;
                         
};





//Constructor() 
Search_algo::Search_algo(char * File_name):file(File_name){
 
              build_vectors();     
                   
}





//Destructor()
Search_algo::~Search_algo () {
      
      // Empty Destructor             
} 


void Search_algo::build_vectors(){
     
     
     int x;
     string str;    
     
     ifstream input;
     input.open(file);
     
     if (!input){
                
         cerr << file <<" not found " << endl;
         exit(1);
      }
     
    
     //Roles_indices
     while (str.compare("#R")!=0 ) getline(input,str);
     //getline(input,str);
     //while(str.empty()) getline(input,str);
     
     input >> Roles_indices;
     
     
     // Perm_indices
     while (str.compare("#P")!=0 ) getline(input,str);

     input >> Perms_indices;   
     
     // D
     
     while (str.compare("#D")!=0) getline(input,str);
      
     getline(input,str);
          
     while (str.compare("#PLB")!=0 ){      
           
           if (!str.empty()){
                             
                   stringstream lineStream1(str);
                   vector<int> v;
                   D.push_back(v);
          
                   while (lineStream1 >> x){
                
                         int last=D.size()-1;
                         D[last].push_back(x);
                   }
          }
          
          getline(input,str);
      
     }
     
     
     // PLB 
     
     while (str.compare("#PLB")!=0 ) getline(input,str);
     getline(input,str);
     while(str.empty()) getline(input,str);

     stringstream lineStream2(str);
          
     while (lineStream2 >> x){
                
           PLB.push_back(x);
     }
     
     
     //PUB_PLB
     
     while (str.compare("#PUB-PLB")!=0 ) getline(input,str);
     getline(input,str);
     while(str.empty()) getline(input,str);
     
     stringstream lineStream3(str);
          
     while (lineStream3 >> x){
                
           PUB_PLB.push_back(x);     
     }
     
     //OBJ
     
     while (str.compare("#OBJ")!=0 )  getline(input,str);
     
     input >> OBJ;
     
     
     // PR_PLB ,  PR_PUB_PLB  and  PR_POUT
     
     input.seekg(0,ios::beg);
      
     while (str.compare("#PR")!=0 ) getline(input,str);
     
     getline(input, str);
          
     while (str.compare("#D")!=0 ){      
           
           if (!str.empty() ){ 
                            
                    stringstream lineStream4(str);
                    lineStream4 >> x;
           
                    if (member(PLB,x)){
                  
                           vector<int> v;
                           PR_PLB.push_back(v); 
                           int last=PR_PLB.size()-1;
                           PR_PLB[last].push_back(x);
                        
                           while (lineStream4 >> x){
                
                                 PR_PLB[last].push_back(x);
                           }
                  
                    }else if( member(PUB_PLB,x) ){
                 
                          vector<int> v;
                          PR_PUB_PLB.push_back(v); 
                          int last=PR_PUB_PLB.size()-1;
                          PR_PUB_PLB[last].push_back(x);
                        
                          while (lineStream4 >> x){
                
                                PR_PUB_PLB[last].push_back(x);
                          }
                                
                     }else {

			   vector<int> v;
                           PR_POUT.push_back(v); 
                           int last=PR_POUT.size()-1;
                           PR_POUT[last].push_back(x);
                        
                           while (lineStream4 >> x){
                
                                 PR_POUT[last].push_back(x);
                           }
		      }


             }
          
           getline(input,str);
      
     }
   
   
   input.close();  
}




bool Search_algo::Is_any(int k){
          
          if (PR_PLB.size()==0) {
                  
                  return false;
          }
                               
          vector<int> roles;
               
          int64_t total_comb=1;
          
          for(int i=0; i<PR_PLB.size(); i++ ){
                 
                 total_comb *= max( int(PR_PLB[i].size()-1), 1);
                 
                 if (total_comb <0) { 
                                
                          cout << "Overflow occured" << endl; 
                          break;
                 }
                 
          }
           
                   
          for(int i=1; i<total_comb+1; i++){
                 
                   roles=gen_roles(i);
     
                   if ( PUB_PLB_OK(roles, k) && D_OK(roles) && POUT_OK(roles) ){
                                               
                        return true;
                   }
                  
          }              
               
     return false;          
}


//

vector<int> Search_algo::gen_roles(int n){
            
            int size=PR_PLB.size();
        
            int Comb_num[size];        
            Comb_num[0]=max(int(PR_PLB[0].size()-1),1);
            
            for(int i=1; i<size; i++ ){
                  
                  Comb_num[i] = max(int(PR_PLB[i].size()-1), 1)*Comb_num[i-1]; 
            }
                         
            vector<int> roles;
            
            int idx= (n-1) % (PR_PLB[0].size()-1)+1;
            int role= PR_PLB[0][idx];
            roles.push_back(role);
            
            for(int i=1; i<size; i++){
              
                    idx = ( int(ceil(double(n)/Comb_num[i-1]))-1) % (PR_PLB[i].size()-1)+1;
                    role= PR_PLB[i][idx];     
                    roles.push_back(role);
            } 
            
 
      return roles;              
                           
}



bool Search_algo::POUT_OK(const vector<int> &roles) const{

	for (int i=0;i<PR_POUT.size(); i++ ){
             
             for (int j=1; j<PR_POUT[i].size();j++){
                 
                 if ( member(roles, PR_POUT[i][j]) ) {
                           
			return false;
                         
                 }
             }
         }

return true;

}



bool Search_algo::PUB_PLB_OK(const vector<int> &roles, int k) const{
     
              
     
         int num_acc=0;  // num of perms in PUB/PLB 's been accessed
         
         for (int i=0;i<PR_PUB_PLB.size(); i++ ){
             
             for (int j=1; j<PR_PUB_PLB[i].size();j++){
                 
                 if ( member(roles, PR_PUB_PLB[i][j]) ) {
                           num_acc++;
                           break;
                 }
             }
         }
         
         
         if (OBJ==0){
                     
              return (num_acc<=k);
              
         }else if(OBJ==1){
               
               return (num_acc>=k);
         }
     
}





bool Search_algo::D_OK(const vector<int> & roles) const{
     
           for(int i=0; i<D.size(); i++){
                   
                   if ( !d_OK( D[i],roles ) ){
                        
                        return false;
                   }
           }     
         
     return true;
}






bool Search_algo::d_OK(const vector<int> &d , const vector<int> & roles) const{
     
           int num_rls_act=0;
           
           for(int i=0; i<d.size()-1; i++){
                   
                   if ( member(roles,d[i])){
                        
                        num_rls_act++;
                   }
           }     
      
         
     int last=d.size()-1;
     return (num_rls_act<d[last]);
     
}



bool Search_algo::member(const vector<int> &v, const int &obj ) const{
                   
        for (int i=0; i<v.size(); i++){
            
            if  (v[i]==obj)  return true;
        }
        
        return false;
        
}


void Search_algo::printVector(const vector<int> &v) const{
     
      for(int i=0; i<v.size() ; i++){

           cout << v.at(i) <<" " ;

      }
}

#endif












