#ifndef POLICY2CNF_H
#define POLICY2CNF_H


#include <iostream>
#include<vector>
#include<string.h>
#include <sstream>
#include<fstream>
#include"stdlib.h"
#include"K_Check.h"


class Session {
      
      private: 
              
              char* Session_file;
              int Roles_indices;                     // 25  I.e., roles are 1 - 25 
              int Perms_indices;                     // 50  I.e., perms are 26 - 75 */
              vector< vector<int> > PR;              // Each perm is assigned to those roles
              vector< vector<int> > D;               // Each Constraint is assigned to those Roles 
              vector<int> PLB;                       // set of permissions, P_lb, that mut be activate 
              vector<int> PUB_PLB;                   // P_ub - P_lb, NOTE: this line contains 0 for P_ub -P_lb = \emptyset
              int k_role, k_perm;                    // k_role = thereshold for number of roles activated,  
                                                     // k_perm= thereshold for the number of activated permissions in PUB\PLB

                              
      public:
             int PUB_PLB_size;
             int free_index;
             vector<int> CNF;
             int OBJ_role ;   
             int OBJ_perm;                // 0 means min, 1 means max. NOTE: this is irrelevant if #PUB-PLB above has 0  
              
              
             Session (char*, int=0, int=0, int=0, int=0);   // Session gets 5 arguments; File name, Roles threshold,
					                    // PUB\PLB thereshold, Objective for roles and Objective 
							    // for permission
             ~Session ();
             
             void build_vectors();
             void copy_to_CNF(const vector<int>& v);
             bool member(const vector<int> &v, const int & obj ) const;
             void generate_plb_CNF(const vector<int> & );
             void generate_pub_plb_CNF(const vector<int> &);
             void generate_perm_ouside_CNF(const vector<int> &);
             void generate_PR_CNF();
             void generate_D_CNF();
             void generate_PUB_CNF();
             void generate_R_CNF();
             void printVector(const vector<int> &) const;
             void printMatrix(const vector< vector<int> > &) const;
             
             void Generate_CNF();
            
             
             
};

//Constructor() 
Session::Session(char * File_name, int n, int m, int obj_role , int obj_perm):Session_file(File_name), k_role(n), 
                                          k_perm(m), OBJ_role(obj_role), OBJ_perm(obj_perm) {
 
              build_vectors();     // Extract information from input file into some vectors
              free_index=Roles_indices+Perms_indices+1;      // free index will be used for CNF variables
              PUB_PLB_size=PUB_PLB.size();    
                     
                   
}



//Destructor()
Session::~Session () {
      
      // Empty Destructor             
} 




void Session::build_vectors(){
     
     
     int x;
     string str;    
     
     ifstream input;
     input.open(Session_file);
     if (!input){
                
         cerr << Session_file <<" not found " << endl;
         exit(1);
      }
     
     
     //Roles_indices
     while (str.compare("#R")!=0 ) getline(input,str);
     
     input >> Roles_indices;
     
     // Perm_indices
     while (str.compare("#P")!=0 ) getline(input,str);
     
     input >> Perms_indices;
     
     // PR 
      
     while (str.compare("#PR")!=0 ) getline(input,str);
     
     getline(input, str);
          
     while (str.compare("#D")!=0 && !str.empty()){      
           
           stringstream lineStream1(str);
           vector<int> v;
           PR.push_back(v);
          
          while (lineStream1 >> x){
                
               int last=PR.size()-1;
               PR[last].push_back(x);
          }
          
          getline(input,str);
      
     }
     
     // D
     
     while (str.compare("#D")!=0) getline(input,str);
      
     getline(input,str);
          
     while (str.compare("#PLB")!=0 && !str.empty()){      
           
           stringstream lineStream2(str);
           vector<int> v;
           D.push_back(v);
          
          while (lineStream2 >> x){
                
               int last=D.size()-1;
               D[last].push_back(x);
          }
          
          getline(input,str);
      
     }
     
     
     // PLB 
     
     while (str.compare("#PLB")!=0 ) getline(input,str);
     
     getline(input,str);
     stringstream lineStream3(str);
          
     while (lineStream3 >> x){
                
           PLB.push_back(x);
     }
     
     
     //PUB_PLB
     
     while (str.compare("#PUB-PLB")!=0 ) getline(input,str);
     
     getline(input,str);
     stringstream lineStream4(str);
          
     while (lineStream4 >> x){
                
           PUB_PLB.push_back(x);
           
           
     }
 
 /*    
     //OBJ
     
     while (str.compare("#OBJ")!=0 ) getline(input,str);
     
     input >> OBJ;
     
   */
     
     input.close();  
}


void Session::Generate_CNF(){
     
        generate_PR_CNF();           // CNF for permission-to-roles assignments
        generate_D_CNF();           // CNF for the constraints in D
        generate_PUB_CNF();         //  CNF for optimizing over PUB\PLB
        generate_R_CNF();           // CNF for optimizing over roles
     
     
}


void Session::generate_R_CNF(){
     
     vector<int> Roles;
     
     for (int i=1; i<= Roles_indices; i++){
         Roles.push_back(i);
     }
     
     if (k_role <= Roles.size() ){ 
                 
                K_Check * checker=new K_Check(Roles, k_role, OBJ_role);
                checker->Generate_CNF(free_index);
          
                copy_to_CNF(checker->CNF);
          
                delete checker;         
     }
     
     
}

void Session::generate_PR_CNF(){
            
                    
           for (int i=0; i<PR.size(); i++) {
              
             
              	if ( member(PLB, PR[i][0]) ){
                             
                    	generate_plb_CNF(PR[i]);

              	}else if (member(PUB_PLB, PR[i][0]) ){
                    
                  	  generate_pub_plb_CNF(PR[i]);

              	}else{ 
                    
                   	 generate_perm_ouside_CNF(PR[i]);
             	}
          }              
}


// CNF for the constraint d of form 
// d =( v , k) where  v is of form  v =  r_1, ..., r_n  or v = p_1, ... p_n

void Session::generate_D_CNF(){
          
          int k, last_pos;
          vector<int> curr_vec;
          
          
          for (int i=0; i<D.size(); i++){
               
                last_pos=D[i].size()-1;
                k=D[i][last_pos];
                k=k-1;   // because of the definition of last element in any row of DA

                D[i].pop_back();
                curr_vec= D[i];
                 
                if (k < curr_vec.size()){  
                            
                      K_Check * checker=new K_Check(curr_vec, k, 0);
                      checker->Generate_CNF(free_index);
          
                      copy_to_CNF(checker->CNF);
                
                      delete checker;
                }
                
          }
          
}


void Session::generate_PUB_CNF(){
          
 /* 
    if ( k_perm == 0 && OBJ_perm == 0){
		
		   for(int i=0; i<PUB_PLB.size(); i++){
	
			       CNF.push_back(-PUB_PLB[i]);
			       CNF.push_back(0); 
            }
	return;

	}

*/

	if (k_perm < PUB_PLB.size() ){ 
                 
                K_Check * checker=new K_Check(PUB_PLB, k_perm, OBJ_perm);
                checker->Generate_CNF(free_index);
          
                copy_to_CNF(checker->CNF);
          
                delete checker;         
     }
      
          
}

void Session::generate_plb_CNF(const vector<int> &v ){
                                         
       for(int i=1; i<v.size(); i++){
               
               CNF.push_back(v[i]);
                                      
       }
       
       CNF.push_back(0); 
                         
                                         
}



void Session::generate_pub_plb_CNF(const vector<int> & v ){
                                         
       for(int i=1; i<v.size(); i++){
               
               CNF.push_back(v[0]);
               CNF.push_back(-v[i]); 
               CNF.push_back(0);                                  
       }
       

       
       CNF.push_back(-v[0]);

       for(int i=1; i<v.size(); i++){

		 CNF.push_back(v.at(i)); 
       }

       CNF.push_back(0);  
                                    
}


void Session::generate_perm_ouside_CNF(const vector<int> & v){
     
         for(int i=1; i<v.size(); i++){
               
              // CNF.push_back(v[0]);
               CNF.push_back(-v[i]); 
               CNF.push_back(0);                                  
       }
       

     
}








bool Session::member(const vector<int> &v, const int &obj ) const{
                   
        for (int i=0; i<v.size(); i++){
            
            if  (v[i]==obj)  return true;
        }
        
        return false;
        
}



void Session::copy_to_CNF(const vector<int> &v){
     
     for (int i=0; i< v.size(); i++){
         
         CNF.push_back(v[i]);
         
     }
     
}


void Session::printVector(const vector<int> &v) const {


      for(int i=0; i<v.size() ; i++){


           cout << v[i] <<" " ;

           if (v[i]==0 ){
                       
               cout << endl;
           }
      }
}



void Session::printMatrix(const vector< vector<int> > &W)const{
     
     for (int i=0 ; i< W.size(); i++){
         
         for (int j=0; j< W.at(i).size() ; j++) {
             
             cout << W[i][j] << " "; 
         }
         
         cout << endl;
     }
    cout <<endl;         
}


#endif



